;+
; NAME:
;   MGHgrBarb
;
; PURPOSE:
;   This class implements a set of wind/current barbs by joining up the
;   appropriate vertices in a single IDLgrPolyline.
;
; CATEGORY:
;   Object graphics.
;
; SUPERCLASS:
;   IDLgrModel
;
; PROPERTIES:
;   Properties not recognised by this class are passed to the barb
;   object (an IDLgrPolyline). The following properties are supported
;   directly
;
;     BARB_COLORS (Init)
;       A vector of indexed or RGB colours (ie an [n] or [3,n] byte
;       array) specifying the colour of each barb.
;
;     DATAU, DATAV, DATAW (Init)
;       These specify the displacement of the tip of each barb from
;       its base in the x, y & z directions respectively. Optional,
;       default = 0.
;
;     DATAX, DATAY, DATAZ (Init)
;       Each of these  is a scalar or vector specifying the location
;       of the base of each barb. Optional, default = 0.
;
;     NORM_SCALE (Init, Get, Set)
;       A 3-element boolean vector specifying how the SCALE property is to
;       be interpreted for each direction. If NORM_SCALE is true for a
;       given direction, then the corresponding SCALE is interpreted
;       as a length in normalised coordinates, otherwise it is
;       interpreted as a length in data coordinates (the same ones as used
;       for DATA[X,Y,Z]).
;
;     SCALE (Init, Get, Set)
;       A 3-element double-precision vector specifying the length of a
;       unit velocity vector. This property can be (and often will be)
;       passed to Init or GetProperty as a scalar. Default is [1,1,1].
;
; TO DO:
;   Direct appropriate keywords (e.g. HIDE) to the model rather than
;   the atom. Separate colour recalculation from dimension
;   recalculation--you don't really need to recalculate all the
;   dimensions when the colour is changed.
;
;###########################################################################
;
; This software is provided subject to the following conditions:
;
; 1.  NIWA makes no representations or warranties regarding the
;     accuracy of the software, the use to which the software may
;     be put or the results to be obtained from the use of the
;     software.  Accordingly NIWA accepts no liability for any loss
;     or damage (whether direct of indirect) incurred by any person
;     through the use of or reliance on the software.
;
; 2.  NIWA is to be acknowledged as the original author of the
;     software where the software is used or presented in any form.
;
;###########################################################################
;
; MODIFICATION HISTORY:
;   Mark Hadfield, 1998-02:
;     Written as MGHgrBarb.
;   Mark Hadfield, 1999-05:
;     Added DATAZ & BARB_COLORS.
;   Mark Hadfield, 2000-07:
;     Substantially rewritten. The meaning of the SCALE keyword has
;     changed so that it is more logical and extensible to the Z
;     dimension--this is NOT BACKWARD-COMPATIBLE. Changed from a
;     subclass of IDLgrPolyline to a subclass of IDLgrModel (with
;     SELECT_TARGET set) in preparation for supporting symbols. Added
;     DATAW property for full 3D support.
;   Mark Hadfield, 2001-07:
;     Updated for IDL 5.5.
;   Mark Hadfield, 2004-07:
;     - Renamed MGHgrBarb.
;     - Handling of scaling changed: the default is now to scale in data
;       coordinates and the NORM_SCALE keyword has been added to support
;       the original normalised scaling.
;   Mark Hadfield, 2007-09:
;     - Fixed bug: single precision calculations used in CalculateDimensions
;       methods when the object's DOUBLE property is set.
;     - Changed default value of DOUBLE property to 1.
;-
function MGHgrBarb::Init, $
     BARB_COLORS=barb_colors, $
     DATAU=datau, DATAV=datav, DATAW=dataw, $
     DATAX=datax, DATAY=datay, DATAZ=dataz, $
     DESCRIPTION=description, HIDE=hide, NAME=name, $
     NORM_SCALE=norm_scale, $
     REGISTER_PROPERTIES=register_properties, $
     SCALE=scale, SYMBOL=symbol, $
     XCOORD_CONV=xcoord_conv, $
     YCOORD_CONV=ycoord_conv, $
     ZCOORD_CONV=zcoord_conv, $
     _REF_EXTRA=extra

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   ;; Initialise the model

   ok = self->IDLgrModel::Init(DESCRIPTION=description, HIDE=hide, $
                               NAME=name, /SELECT_TARGET)
   if ~ ok then $
        message, BLOCK='MGH_MBLK_MOTLEY', NAME='MGH_M_INITFAIL', 'IDLgrModel'

   ;; Create child objects

   self.barb_atom = obj_new('IDLgrPolyline', /DOUBLE, _STRICT_EXTRA=extra)
   self->Add, self.barb_atom

   self.symbol_atom = obj_new('IDLgrPolyline', LINESTYLE=6)
   self->Add, self.symbol_atom

   ;; Set defaults

   self.norm_scale = 0

   self.scale = 1

   self.xcoord_conv = [0,1]
   self.ycoord_conv = [0,1]
   self.zcoord_conv = [0,1]

   ;; Pass all remaining arguments to SetProperty. (So why not pass
   ;; them all via extra?)

   self->SetProperty, BARB_COLORS=barb_colors, $
        DATAU=datau, DATAV=datav, DATAW=dataw, $
        DATAX=datax, DATAY=datay, DATAZ=dataz, $
        NORM_SCALE=norm_scale, $
        SCALE=scale, SYMBOL=symbol, $
        XCOORD_CONV=xcoord_conv, YCOORD_CONV=ycoord_conv, $
        ZCOORD_CONV=zcoord_conv, _STRICT_EXTRA=extra

   if keyword_set(register_properties) then begin

      self->RegisterProperty, 'NAME', NAME='Name', /STRING
      self->RegisterProperty, 'DESCRIPTION', NAME='Description', /STRING
      self->RegisterProperty, 'LINESTYLE', NAME='Line style', /LINESTYLE
      self->RegisterProperty, 'THICK', NAME='Line thickness', /THICKNESS

   endif

   return, 1

end

; MGHgrBarb::Cleanup
;
pro MGHgrBarb::Cleanup

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   compile_opt LOGICAL_PREDICATE

   ptr_free, self.barb_colors

   ptr_free, self.datax
   ptr_free, self.datay
   ptr_free, self.dataz

   ptr_free, self.datau
   ptr_free, self.datav
   ptr_free, self.dataw

   self->IDLgrModel::Cleanup

end

; MGHgrBarb::GetProperty
;
pro MGHgrBarb::GetProperty, $
     DATAU=datau, DATAV=datav, DATAW=dataw, $
     DATAX=datax, DATAY=datay, DATAZ=dataz, $
     DESCRIPTION=description, HIDE=hide, NAME=name, $
     NORM_SCALE=norm_scale, $
     PARENT=parent, SCALE=scale, SYMBOL=symbol, $
     XCOORD_CONV=xcoord_conv, $
     YCOORD_CONV=ycoord_conv, $
     ZCOORD_CONV=zcoord_conv, _REF_EXTRA=extra

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   self->IDLgrModel::GetProperty, $
        DESCRIPTION=description, HIDE=hide, NAME=name, PARENT=parent

   self.barb_atom->GetProperty, _STRICT_EXTRA=extra

   if arg_present(datax) then $
        datax = ptr_valid(self.datax) ? *self.datax : 0

   if arg_present(datay) then $
        datay = ptr_valid(self.datay) ? *self.datay : 0

   if arg_present(dataz) then $
        dataz = ptr_valid(self.dataz) ? *self.dataz : 0

   if arg_present(datau) then $
        datau = ptr_valid(self.datau) ? *self.datau : 0

   if arg_present(datav) then $
        datav = ptr_valid(self.datav) ? *self.datav : 0

   if arg_present(dataw) then $
        dataw = ptr_valid(self.dataw) ? *self.dataw : 0

   norm_scale = self.norm_scale

   scale = self.scale

   xcoord_conv = self.xcoord_conv
   ycoord_conv = self.ycoord_conv
   zcoord_conv = self.zcoord_conv

   if arg_present(symbol) then self.symbol_atom->GetProperty, SYMBOL=symbol

end

; MGHgrBarb::SetProperty
;
pro MGHgrBarb::SetProperty, $
     BARB_COLORS=barb_colors, $
     DATAU=datau, DATAV=datav, DATAW=dataw, $
     DATAX=datax, DATAY=datay, DATAZ=dataz, $
     DESCRIPTION=description, HIDE=hide, NAME=name, $
     NORM_SCALE=norm_scale, $
     SCALE=scale, SYMBOL=symbol, $
     XCOORD_CONV=xcoord_conv, $
     YCOORD_CONV=ycoord_conv, $
     ZCOORD_CONV=zcoord_conv, _REF_EXTRA=extra

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   self->IDLgrModel::GetProperty, $
        DESCRIPTION=description, HIDE=hide, NAME=name

   self.barb_atom->SetProperty, _STRICT_EXTRA=extra

   recalc = 0B

   if n_elements(barb_colors) gt 0 then begin
      recalc = 1B
      ptr_free, self.barb_colors
      self.barb_colors = ptr_new(barb_colors)
   endif

   if n_elements(datax) gt 0 then begin
      recalc = 1B
      ptr_free, self.datax
      self.datax = ptr_new(datax)
   endif

   if n_elements(datay) gt 0 then begin
      recalc = 1B
      ptr_free, self.datay
      self.datay = ptr_new(datay)
   endif

   if n_elements(dataz) gt 0 then begin
      recalc = 1B
      ptr_free, self.dataz
      self.dataz = ptr_new(dataz)
   endif

   if n_elements(datau) gt 0 then begin
      recalc = 1B
      ptr_free, self.datau
      self.datau = ptr_new(datau)
   endif

   if n_elements(datav) gt 0 then begin
      recalc = 1B
      ptr_free, self.datav
      self.datav = ptr_new(datav)
   endif

   if n_elements(dataw) gt 0 then begin
      recalc = 1B
      ptr_free, self.dataw
      self.dataw = ptr_new(dataw)
   endif

   if n_elements(norm_scale) gt 0 then begin
      recalc = 1B
      case n_elements(norm_scale) of
         1: self.norm_scale = norm_scale[0]   ;;; Assign single value to all 3 elements
         else: self.norm_scale = norm_scale   ;;; Assign vector to vector
      endcase
   endif

   if n_elements(scale) gt 0 then begin
      recalc = 1B
      case n_elements(scale) of
         1: self.scale = scale[0]   ;;; Assign single value to all 3 elements
         else: self.scale = scale   ;;; Assign vector to vector
      endcase
   endif

   if n_elements(symbol) gt 0 then begin
      recalc = 1B
      self.symbol_atom->SetProperty, SYMBOL=symbol
   endif

   if n_elements(xcoord_conv) gt 0 then begin
      recalc = 1B
      self.xcoord_conv = xcoord_conv
   endif

   if n_elements(ycoord_conv) gt 0 then begin
      recalc = 1B
      self.ycoord_conv = ycoord_conv
   endif

   if n_elements(zcoord_conv) gt 0 then begin
      recalc = 1B
      self.zcoord_conv = zcoord_conv
   endif

   if recalc then self->CalculateDimensions

end

; MGHgrBarb::CalculateDimensions
;
pro MGHgrBarb::CalculateDimensions

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   ;; Get data and/or provide defaults.

   datax = ptr_valid(self.datax) ? *self.datax : 0
   datay = ptr_valid(self.datay) ? *self.datay : 0

   datau = ptr_valid(self.datau) ? *self.datau : 0
   datav = ptr_valid(self.datav) ? *self.datav : 0

   use_z = ptr_valid(self.dataz) || ptr_valid(self.dataw)

   if use_z then begin
      dataz = ptr_valid(self.dataz) ? *self.dataz : 0
      dataw = ptr_valid(self.dataw) ? *self.dataw : 0
   endif

   ;; Calculate number of barbs

   n_barbs = $
            n_elements(datax) > n_elements(datay) > $
            n_elements(datau) > n_elements(datav)

   if use_z then $
        n_barbs = n_barbs > n_elements(dataz) > n_elements(dataw)

   ;; Calculate velocity scale in data coordinates

   data_scale = self.scale
   if self.norm_scale[0] then data_scale[0] /= self.xcoord_conv[1]
   if self.norm_scale[1] then data_scale[1] /= self.ycoord_conv[1]
   if self.norm_scale[2] then data_scale[2] /= self.zcoord_conv[1]

   ;; Create & fill a polyline vertex array.

   self->GetProperty, DOUBLE=double

   case use_z of

      0: begin
         vert = make_array(2, 2*n_barbs, DOUBLE=double)
         vert[4*lindgen(n_barbs)  ] = datax
         vert[4*lindgen(n_barbs)+1] = datay
         vert[4*lindgen(n_barbs)+2] = datax + datau*data_scale[0]
         vert[4*lindgen(n_barbs)+3] = datay + datav*data_scale[1]
      end

      1: begin
         vert = make_array(3, 2*n_barbs, DOUBLE=double)
         vert[6*lindgen(n_barbs)  ] = datax
         vert[6*lindgen(n_barbs)+1] = datay
         vert[6*lindgen(n_barbs)+2] = dataz
         vert[6*lindgen(n_barbs)+3] = datax + datau*data_scale[0]
         vert[6*lindgen(n_barbs)+4] = datay + datav*data_scale[1]
         vert[6*lindgen(n_barbs)+5] = dataz + dataw*data_scale[2]
      end

   endcase

   ;; Handle missing vertex values. An IDLgrPolyline does not permit
   ;; non-finite vertices so set them to 0 (the associated line segments
   ;; will be omitted from the connectivity array.

   missing = where(~ finite(vert), n_missing_vertices)
   if n_missing_vertices gt 0 then vert[missing] = 0

   ;; Specify connections between polyline vertices.

   conn = lonarr(3*n_barbs)

   case n_missing_vertices gt 0 of

      0: begin
         conn[3*lindgen(n_barbs)  ] = 2
         conn[3*lindgen(n_barbs)+1] = 2*lindgen(n_barbs)
         conn[3*lindgen(n_barbs)+2] = 2*lindgen(n_barbs)+1
      end

      1: begin
         case use_z of
            0: good = where(replicate(1, n_barbs) and finite(datax) and $
                            finite(datay) and finite(datau) and $
                            finite(datav), n_good_data)
            1: good = where(replicate(1, n_barbs) and finite(datax) and $
                            finite(datay) and finite(dataz) and $
                            finite(datau) and finite(datav) and $
                            finite(dataw), n_good_data)
         endcase
         for g=0,n_good_data-1 do begin
            i = good[g]
            conn[3*g]   = 2
            conn[3*g+1] = 2*i
            conn[3*g+2] = 2*i+1
         endfor
         conn[3*n_good_data] = -1 ; Terminate the list of polylines
      end

   endcase

   ;; Barbs will be coloured via VERT_COLORS

   if ptr_valid(self.barb_colors) then begin

      barb_colors = *(self.barb_colors)

      dims = size(barb_colors, /DIMENSIONS)
      case size(barb_colors, /N_DIMENSIONS) of
         1: begin
            vert_colors = bytarr(2, dims[0])
            vert_colors[0,*] = barb_colors
            vert_colors = reform(vert_colors, 2*dims[0])
         end
         2: begin
            if dims[0] ne 3 then $
                 message, 'The inner dimension of BARB_COLORS must be 3'
            vert_colors = bytarr(3, 2, dims[1])
            vert_colors[*,0,*] = barb_colors
            vert_colors = reform(vert_colors, 3, 2*dims[1])
         end
      endcase

   endif else begin

      vert_colors = -1

   endelse

   ;; Set up barb polyline.

   self.barb_atom->SetProperty, DATA=vert, POLYLINES=conn, $
        VERT_COLORS=vert_colors

   ;; Set up symbol polyline (if any) using a subset of the barb
   ;; polyline vertices.  If a symbol is added then deleted, then the
   ;; symbol polyline vertices hang around but they're invisible.

   self.symbol_atom->GetProperty, SYMBOL=symbol

   if n_elements(symbol) gt 0 then $
        self.symbol_atom->SetProperty, DATA=vert[*,2*lindgen(n_barbs)]

   ;; Scale & translate the model

   self->IDLgrModel::Reset

   self->IDLgrModel::Scale, $
        self.xcoord_conv[1], self.ycoord_conv[1], self.zcoord_conv[1]

   self->IDLgrModel::Translate, $
        self.xcoord_conv[0], self.ycoord_conv[0], self.zcoord_conv[0]

end


; MGHgrBarb__Define

pro MGHgrBarb__Define

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   struct_hide, $
        {MGHgrBarb, inherits IDLgrModel, $
         barb_atom: obj_new(), barb_colors: ptr_new(), $
         symbol_atom: obj_new(), scale: dblarr(3), $
         norm_scale: bytarr(3), $
         datax: ptr_new(), datay: ptr_new(), dataz: ptr_new(), $
         datau: ptr_new(), datav: ptr_new(), dataw: ptr_new(), $
         xcoord_conv: dblarr(2), $
         ycoord_conv: dblarr(2), $
         zcoord_conv: dblarr(2)}

end

